clear
echo on
/*
=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=
|     H2 output feedback control problem      |
=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=

  The problem we want to solve is

    min   trace(W)

    s.t.  | Y     I                 B1           |
          |                                      |
          | I     X                 X B1 + F D21 | > 0
          |                                      |
          | B1'   B1' X + D21' F'   W            |

          | A Y + Y A' + B2 L + L' B2'   Y C1' + L' D12' |
          |                                              | < 0
          | C1 Y + D12 L                -I               |

          A' X + X A + C2' F' + F C2 + C1' C1 < 0

                  
  where X, F, Y, L and W are the optimization variables.

=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=
*/

pause, home

// Set the output display step and required precision
display = 5;
eps = 1e-4;

// Define the data
A = [-2 1 1 1; 3 0 0 2; -1 0 -2 -3; -2 -1 2 -1];
B1 = [1 0 0; 0 0 0; 1 0 0; 0 0 0];
B2 = [0 0; 1 0; 0 0; 0 1];
C1 = [1 0 -1 0; 0 0 0 0; 0 0 0 0];
C2 = [1 0 0 0; 0 0 1 0];
D12 = [0 0; 1 0; 0 1];
D21 = [0 1 0; 0 0 1];
n = size(A,1);
m = size(B2,2);
p = size(C2,1);

pause, home

// Define the variables
var sym X(n) F(n,p) sym Y(n) L(m,n) sym W(size(B1'*B1))

// and the inequalities
lmi = block( Y; 
             eye(n) X;
             B1'    B1'*X+D21'*F' W )

lmi = block( -sym(A*Y+B2*L);
             -(C1*Y+D12*L)   eye(size(C1*C1')) )

lmi = -( sym(X*A+F*C2)+C1'*C1 )

obj = trace(W);

pause, home

// Solve
solve
X, F, Y, L, W

pause, home

/*
  Calculate the controller:

    U V' = I - X Y

    Z = A + | Y   L' | | A   B2 |' | X  |
                       | C2  0  |  | F' |

          -1       -1        -1              -1
    Ac = U   M' (V'),  Bc = U   F, Cc = L (V')
    
*/
V = Y;
U = inv(Y) - X
Z = A + [Y L']*[A B2; C2 zeros(size(C2*B2))]'*[X; F']
M = -Z-(Y*C1'+L'*D12')*C1
Cc = L / V'
Bc = U \ F
Ac = U \ M' / V

pause, home

// and check the stability of the closed loop system
Af = [A B2*Cc; Bc*C2 Ac];
e = eig(Af);
e(:,1)'

pause, home

/*
=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=
|     Guaranteed cost H2 output feedback      |
|     control problem                         |
=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=

  The problem we want to solve is

    min   trace(W)

    s.t.  | Y     I                 B1           |
          |                                      |
          | I     X                 X B1 + F D21 | > 0
          |                                      |
          | B1'   B1' X + D21' F'   W            |

| Ai Y + Y Ai' + B2i L + L' B2i'   *                                * |
|                                                                     |
| Zi' - M'                         Ai' X + X Ai + C2i' F' + F C2i   * | < 0,
|                                                                     |
| C1 Y + D12 L                     C1                              -I |

             i = 1, ..., N

  where X, F, Y, L, M and W are the optimization variables,
  and Zi are nonlinear matrix functions in the form
  
    Zi = Ai + | Y   L' | | Ai   B2i |' | X  |     
                         | C2i  0   |  | F' |

=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=
*/
pause, home

/*

  We iteratively search a solution for that nonlinear matrix
  inequality problem via the Cross Decomposition Algorithm
  
  
    while ( (f2 - f1) / f2 ) > eps {


      f1 := min  trace( W )
    
            s.t. RXF(X, F, P, S, W)
	  
	  
      f2 := min  trace( W )
    
            s.t. RYL(Q, S, Y, L, W)

    }
  
  where RXF(.) and RYL(.) are LMI which will be defined in the sequel.
  
*/
pause, home

/*

  The LMI RXF is given by

    | P      P                 P B1         |
    |                                       |
    | P      X                 X B1 + F D21 | > 0
    |                                       |
    | B1' P  B1' X + D21' F'   W            |

    | Aki' P + P Aki               *                                * |
    |                                                                 |
    | Ai' P + X Aki + F C2i - S'   Ai' X + X Ai + F C2i + C2i' F'   * | < 0,
    |                                                                 |
    | Ck                           C1                              -I |

             i = 1, ..., N

  where X, F, P, S and W are the optimization variables and

                         -1                              -1
    Aki := Ai + B2i Lk Yk  ,         Ck := C1 + D12 Lk Yk
    
*/
pause, home

/*

  The LMI RYL is given by

    | Y     Q     B1 |
    |                |
    | Q     Q     Bk | > 0
    |                |
    | B1'   Bk'   W  |

    | Ai Y + Y Ai' + B2i L + L' B2i'   *                * |
    |                                                     |
    | Q Ai' + Aki Y + B2i L - S'       Aki Q + Q Aki'   * | < 0,
    |                                                     |
    | C1 Y + D12 L                     C1 Q            -I |

             i = 1, ..., N

  where Q, S, Y, L and W are the optimization variables and

                  -1                              -1
    Aki := Ai + Xk   Fk C2i,         Bk := B1 + Xk   Fk D21
  
*/
pause, home

/*

  As in the state feedback example we try to find an output
  feedback controller that keeps stability and performance
  even in the occurrence of some sensor or actuator failure.
  For the same data as before we define
  
*/
B21 = [0 0; 1 0; 0 0; 0 1];  // All actuators are ok
B22 = [0 0; 0 0; 0 0; 0 1];  // First actuator failure
B23 = [0 0; 1 0; 0 0; 0 0];  // Second actuator failure

C21 = [1 0 0 0; 0 0 1 0];    // All sensors are ok
C22 = [0 0 0 0; 0 0 1 0];    // First sensor failure
C23 = [1 0 0 0; 0 0 0 0];    // Second sensor failure

Ax  = [A;   A;   A;   A;   A];
B2x = [B21; B22; B23; B21; B21];
C2x = [C21; C21; C21; C22; C23];
//      ^    ^    ^    ^    ^
//      |    |    |    |    |
//     Ok    |    |    |    Sensor 2 fails
//           |    |   Sensor 1 fails
//           |   Actuator 2 fails
//         Actuator 1 fails

pause, home

Af = [A B22*Cc; Bc*C21 Ac];
e = eig(Af);
e(:,1)'
// It is worth noticing that the controller design of the last
// does not posses that reliability property, which can be seen
// from the eigenvalues of the closed loop matrix Af built 
// considering the pair (B22,C21)

pause, home

// Define the necessary condition initialization problem
echo off
exec ph2init
echo on

// solve it
solve;

// and set the initial solution
Yk = Y; Lk = L;
pause, home

// Cross decomposition algorithm
prec = 1e-2; k = 1;
while 1
        if (k == 1), pause, home, end
        if (k > 1), echo off, end

// Define RXF
	exec ph2XF
        if (k == 1), pause, home, end

// Solve it
	solve;

        echo on
	Xk = X; Fk = F;
	f1 = eval(obj)
        if (k == 1), pause, home, end

        if (k > 1), echo off, end

// Define RYL
	exec ph2YL
        if (k == 1), pause, home, end

// Solve it
	solve;

        echo on
	Yk = Y; Lk = L;
	f2 = eval(obj)
        if (k == 1), pause, home, end

	if ( (f1 - f2) / f2 < prec )
		break;
	end
	k = k + 1;
end

pause, home

echo on
/*
  Calculate the controller:

           -1              -1           -1
    M = S Q,   F = ( Xk Q )   Fk,  X = Q

    U V' = I - X Y
             -1                -1
    U = X = Q  ,          V = X   - Y = Q - Y

          -1       -1      -1           -1
    Ac = U   M' (V')  = Q Q   S' (Q - Y)
    
          -1               -1        -1
    Bc = U   F = Q ( Xk Q )   Fk = Xk   Fk
    
	       -1            -1
    Cc = L (V')   = L (Q - Y)
    
*/
Cc = L /  (Q - Y);
Bc = Xk \ Fk;
Ac = S' / (Q - Y);

pause, home

// and check the stability of the closed loop system
v = e = [];
for i = 1:5
        Afi = [A B2x((i-1)*n+1:i*n,:)*Cc; Bc*C2x((i-1)*p+1:i*p,:) Ac];
	e = eig(Afi);
	v = [v; e(:,1)];
end
v'
