clear
echo on
/*
=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=
|      H2 state feedback control problem      |
=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=

  The problem we want to solve is

    min   trace(W)

    s.t.  | Y              Y C1' + L' D12' |
          |                                | > 0
          | C1 Y + D12 L   W               |
               
          A Y + Y A' + B2 L + L' B2' + B1 B1' < 0

  where Y, L and W are the optimization variables.

=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=
*/

pause, home

clear

// Define the matrices
A = [-2 1 1 1; 3 0 0 2; -1 0 -2 -3; -2 -1 2 -1];
B1 = [1 0 0; 0 0 0; 1 0 0; 0 0 0];
B2 = [0 0; 1 0; 0 0; 0 1];
C1 = [1 0 -1 0; 0 0 0 0; 0 0 0 0];
D12 = [0 0; 1 0; 0 1];
D21 = [0 1 0; 0 0 1];
n = size(A,1);
m = size(B2,2);

// Define the variables
var sym Y(n) L(m,n)
var sym W(size(C1*C1'))

// Y is square symmetric n x n,
// L is m x n and
// W is square symmetric of size(C1*C1').

// Set the required precision
eps = 1e-4;
pause, home

// Define the inequalities
lmi = block(Y; 
            C1*Y+D12*L  W)
lmi = -(A * Y + Y * A' + B2 * L + L' * B2' + B1 * B1')
obj = trace(W)

pause, home

// Solve and look at the solution
solve;
Y, L, W

pause, home

//                                    -1
// Calculate the controller => K = L Y
K = L / Y

// Check if the closed loop system is stable
Af = A + B2 * K;   // Dynamic system closed loop matrix
eig(Af)            // Get the eigenvalues of Af

/*

  Notice that as LmiSolver does not handle complex matrices
  the eigenvalues are returned as a two column matrix where
  the first and second columns correspond, respectively, to
  the real and imaginary part of the eigenvalues.
  
*/

pause, home

/*
=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=
|      Decentralized H2 state feedback        |
|      control problem                        |
=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=

  We want to solve the same problem as before
  
    min   trace(W)

    s.t.  | Y              Y C1' + L' D12' |
          |                                | > 0
          | C1 Y + D12 L   W               |

          A Y + Y A' + B2 L + L' B2' + B1 B1' < 0
                  
  where Y, L and W are the optimization variables, 
  subject to the following decentralization 
  constraint
  
           -1  | K1  0 |
    K = L Y  = |       |
               |  0 K2 |
  
  that provides the decentralized control input
  
         | u1 |           | x1 |    | K1 x1 |
    u =  |    | = K x = K |    | =  |       | 
         | u2 |           | x2 |    | K2 x2 |
    
=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=
*/

pause, home

/*
  Considering the same data as before,
  redefine Y as a nonsymmetric matrix just for fun!
*/
clear Y, var Y(n)

// Impose a symmetric Y mask!
mask(Y) = [ [1 2; 2 3] zeros(2,2); zeros(2,2) [4 5; 5 6] ]

// and the desired L mask
mask(L) = [ones(1,2) zeros(1,2); zeros(1,2) ones(1,2)]

pause, home

solve;
// Solve and get the controller
Y, L
K = L / Y

pause, home

/*
=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=
|      Guaranteed cost H2 state feedback      |
|      control problem                        |
=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=

  The problem we want to solve is

    min   trace(W)

    s.t.  | Y              Y C1' + L' D12' |
          |                                | > 0
          | C1 Y + D12 L   W               |
	  
          Ai X + X Ai' + B2i Y + Y' B2i' + B1 B1' < 0,
	     i = 1 ,..., N
                
  where Y, L and W are the optimization variables.

=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=
*/

pause, home

/* 
  In our example we want to design a state feedback 
  controller which keeps system stability and performance
  even in the occurrence of some actuator failure. For the
  same data we define the variables and parameter matrices.
*/
clear lmi var
var sym Y(n) L(m,n) sym W(size(C1*C1'))

// Define all possible actuator states
B21 = [0 0; 1 0; 0 0; 0 1];  // Everything is ok
B22 = [0 0; 0 0; 0 0; 0 1];  // First actuator failure
B23 = [0 0; 1 0; 0 0; 0 0];  // Second actuator failure

B2x = [B21; B22; B23];

// define the cost lmi
lmi = block(Y; 
            C1*Y+D12*L  W);
	    
pause, home

/*
  Then we define the inequalities associated with matrices B2i,
  i = 1, 2, 3 with the help of a for loop
*/
for i = 1:3
	lmi = -(sym(A * Y + B2x((i-1)*n+1:i*n,:) * L) + B1*B1');
end

pause, home

// Let's take a look at the inequalities we have defined
lmi

pause, home

/*
  Notice that the values of the for variable 'i' have
  been replaced by its actual value in the 'lmi'. This
  is necessary since 'i' does not have any meaning outside
  the 'for' loop. To make things clearer we add another
  inequality as an example
*/
lmi = -(sym(A * Y + B2x((i-1)*n+1:i*n,:) * L) + B1*B1')

pause, home

/*
  In this case 'i' is not a for variable and is not replaced 
  by its value. That feature can be assigned to any scalar
  in LmiSolver via 'inline'. For instance, consider
*/
inline n
lmi = -(sym(A * Y + B2x((i-1)*n+1:i*n,:) * L) + B1*B1')

/*
  Notice that 'for' variables are always 'inline'd. You can
  disable the inline feature with the help of the 'inline off'
  command. The syntax is the following.
*/
inline off n

pause, home

// Finally, clean the workspace
clear lmi(6) lmi(5)

// Define the objective
obj = trace(W)
pause, home

// and solve the problem
solve;
Y, L, W

pause, home

// Calculating the controller
K = L / Y

// We verify that
v = e = [];
for i = 1:3
	e = eig(A + B2x((i-1)*n+1:i*n,:) * K);
	v = [v; e(:,1)];
end
v'

// it stabilizes the system in all proposed situations!

pause, home

/*
=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=
|      Mixed H2 / Hinf state feedback         |
|      control problem                        |
=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=

  The problem we want to solve is

    min   trace(W)

    s.t.  | Y              Y C0' + L' D0' |
          |                               | > 0
          | C0 Y + D0 L   W               |
               
          | A Y + Y A' + B2 L + L' B2' + B1 B1'   Y C1' + L' D12' |
          |                                                       | < 0
          | C1 Y + D12 L                         -gamma I         |

  where Y, L and W are the optimization variables.

=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=
*/

pause, home

// Define the matrices
clear
A = [-2 1 1 1; 3 0 0 2; -1 0 -2 -3; -2 -1 2 -1];
B1 = [1 0 0; 0 0 0; 1 0 0; 0 0 0];
B2 = [0 0; 1 0; 0 0; 0 1];
C0 = [1 1 0 0; 0 0 0 0; 0 0 0 0];
C1 = [1 0 -1 0; 0 0 0 0; 0 0 0 0];
D0 = [0 0; 1 0; 0 1];
D12 = [0 0; 1 0; 0 1];
D21 = [0 1 0; 0 0 1];
gamma = 2;
n = size(A,1);
m = size(B2,2);

// Define the variables
var sym Y(n) L(m,n) sym W(size(C0*C0'))

// Define the inequalities
lmi = block(Y; 
            C0*Y+D0*L  W);
lmi = block( -(sym(A*Y+B2*L)+B1*B1');
             -(C1*Y+D12*L)   gamma*eye(size(C1*C1')) )
obj = trace(W)

pause, home

solve;
